//-----------------------------------------------------------------------------
// Wrapper class for using Windows Services
// 
// Copyright (c) 2001/2002 by Dean Harding <dean@codeka.com>
//-----------------------------------------------------------------------------
#include <windows.h>
#include <string>
#include <iostream>
#include "service.h"

using std::cout;
using std::endl;

//-----------------------------------------------------------------------------
CService CService::m_handle;

//-----------------------------------------------------------------------------
CService::CService()
: m_quitEvent( NULL ), m_name( "MyService" )
{
}
//-----------------------------------------------------------------------------
CService::~CService()
{
}

//-----------------------------------------------------------------------------
void CService::Execute( int numArgs, char **args )
{
	Action action = Unknown;
	int firstArg;
	std::string user, password;

	if( numArgs >= 3 )
	{
		firstArg = 3;

		// we first assume it's one of the control arguments, otherwise it
		// might be starting from the SCM with two arguments
		if( std::string( args[1] ) == std::string( "-install" ) )
			action = DoInstall;
		else if( std::string( args[1] ) == std::string( "-uninstall" ) )
			action = DoUninstall;
		else if( std::string( args[1] ) == std::string( "-start" ) )
			action = DoStart;
		else if( std::string( args[1] ) == std::string( "-stop" ) )
			action = DoStop;
		else if( std::string( args[1] ) == std::string( "-console" ) )
			action = DoConsole;
		else
		{
			action = DoRun;
			firstArg = 1;
		}

		if( action != DoRun )
			m_name = args[2];
		else
			m_name = args[0];
	}
	else
	{
		// if we have a funny number of arguments, then we assume that we're
		// being called from the SCM applet
		firstArg = 1;
		action = DoRun;
		m_name = args[0];
	}

	// if we're installing, check to see if there's a username/password
	if( action == DoInstall && numArgs == 5 )
	{
		user = args[3];
		password = args[4];
	}
	// otherwise, we only accept more than 3 args if we're running or console
	else if( action != DoRun && action != DoConsole && numArgs != 3)
	{
		ShowUsage();
		return;
	}

	switch( action )
	{
	case DoInstall:
		Install( user, password );
		return;
	case DoUninstall:
		Uninstall();
		return;
	case DoStart:
		Start();
		return;
	case DoStop:
		Stop();
		return;
	case DoConsole:
		{
			// set up the args array.  Since we want the Run() method to
			// think it's being run from the control panel applet, we have
			// to do it like this.
			std::string *newArgs = new std::string[numArgs - firstArg + 1];

			newArgs[0] = m_name;
			for( int i = 1; i < numArgs - firstArg; i++ )
				newArgs[i] = args[i + firstArg - 1];

			Console( numArgs - firstArg + 1, newArgs );
		}
		return;
	}

	// if we get here, then we assume we're DoRun, and are being started from the
	// SCM applet.
	SERVICE_TABLE_ENTRY dispatchTable[] = 
	{
		{ m_name.c_str(), &ServiceDispatch },
		{ NULL, NULL }
	};

	if( ::StartServiceCtrlDispatcher( dispatchTable ) == 0 )
	{
		// if this fails, it's probably because someone started us from
		// the command line.  Print a message telling them the "usage"

		ShowUsage();
	}
}

//-----------------------------------------------------------------------------
void WINAPI CService::ServiceDispatch( DWORD numArgs, char **args )
{
	CService *service = GetHandle();
	service->m_name = args[0];

	// we have to initialse the service-specific stuff
	memset( &service->m_status, 0, sizeof(SERVICE_STATUS) );
	service->m_status.dwServiceType = SERVICE_WIN32;
	service->m_status.dwCurrentState = SERVICE_START_PENDING;
	service->m_status.dwControlsAccepted = SERVICE_ACCEPT_STOP;

	service->m_statusHandle = ::RegisterServiceCtrlHandler( service->m_name.c_str(), &ServiceCtrlHandler );
	if( service->m_statusHandle == (SERVICE_STATUS_HANDLE)0 )
	{
		DisplayError( "Failed to register service control handler" );
		return;
	}

	// copy the arguments to an ArmsString array
	std::string *newArgs = new std::string[numArgs];
	for( DWORD i = 0; i < numArgs; i++ )
		newArgs[i] = args[i];

	if( !service->Run( numArgs - 1, newArgs ) )
	{
		service->Shutdown();

		service->m_status.dwCurrentState = SERVICE_STOPPED;
		service->m_status.dwCheckPoint = 0;
		service->m_status.dwWaitHint = 0;
	}
	else
	{
		service->m_status.dwCurrentState = SERVICE_RUNNING;
		service->m_status.dwCheckPoint = 0;
		service->m_status.dwWaitHint = 0;
	}

	::SetServiceStatus( service->m_statusHandle, &service->m_status );
}

//-----------------------------------------------------------------------------
void WINAPI CService::ServiceCtrlHandler( DWORD control )
{
	CService *service = GetHandle();

	switch( control )
	{
	case SERVICE_CONTROL_SHUTDOWN:
	case SERVICE_CONTROL_STOP:
		service->Shutdown();

		service->m_status.dwCurrentState = SERVICE_STOPPED;
		service->m_status.dwWin32ExitCode = 0;
		service->m_status.dwCheckPoint = 0;
		service->m_status.dwWaitHint = 0;
		break;
	case SERVICE_CONTROL_INTERROGATE:
		// just set the current state to whatever it is...
		break;
	}

	::SetServiceStatus( service->m_statusHandle, &service->m_status );
}

//-----------------------------------------------------------------------------
bool CService::Run( int numArgs, std::string *args )
{
	// do whatever here.  this method should return once initalization
	// is complete.  Usually, you'll start another thread to do the
	// actual processing of whatever...
	return true;
}

//-----------------------------------------------------------------------------
void CService::Shutdown()
{
	// the service is about to stop, so other threads should be terminated.
	// don't spend too long in this, because this could be as a result of
	// shutting the computer down.  if it doesn't get a response in 2 seconds,
	// windows will kill your process in this case...
}

//-----------------------------------------------------------------------------
void CService::Install( std::string const &user, std::string const &password )
{
	// get a handle to the service control manager
	SC_HANDLE handle = ::OpenSCManager( NULL, NULL, SC_MANAGER_CREATE_SERVICE );
	if( handle == NULL )
	{
		DisplayError( "Could not connect to SCM dataase" );
		return;
	}

	char buffer[256];
	char filePath[256];
	char *dependancies = NULL;
	char const *puser = NULL;
	char const *ppassword = NULL;

	// find out the filename of ourselves so we can install it to the
	// service control manager.
	::GetModuleFileName( NULL, buffer, 256 );
	sprintf( filePath, "\"%s\"", buffer );

	// if there was a username given, then we need to send it to the
	// CreateService method
	if( user.length() > 0 )
	{
		puser = user.c_str();
		ppassword = password.c_str();
	}

	// if this service depends on others being started first,  set
	// dependancies to a \0-separated string of service names

	// create the service!
	SC_HANDLE service = ::CreateService(
		handle,
		m_name.c_str(),
		m_name.c_str(),
		GENERIC_READ | GENERIC_EXECUTE,
		SERVICE_WIN32_OWN_PROCESS,
		SERVICE_AUTO_START,
		SERVICE_ERROR_IGNORE,
		filePath,
		NULL,
		NULL,
		dependancies,
		puser,
		ppassword
	);

	if( service == NULL )
	{
		if( ::GetLastError() != ERROR_SERVICE_EXISTS )
			DisplayError( "Could not create new service.  Check the username and password, then try again." );
	}
	else
	{
		cout << "Service installed!" << endl;
	}

	::CloseServiceHandle( service );
	::CloseServiceHandle( handle );
}
//-----------------------------------------------------------------------------
void CService::Uninstall()
{
	// get a handle to the SCM
	SC_HANDLE handle = ::OpenSCManager( NULL, NULL, SC_MANAGER_ALL_ACCESS );
	if( handle == NULL )
	{
		DisplayError( "Could not connect to SCM dataase" );
		return;
	}

	// open a handle to the service
	SC_HANDLE service = ::OpenService( handle, m_name.c_str(), DELETE );
	if( service == NULL )
	{
		::CloseServiceHandle( handle );
		return;
	}

	// remove the service!
	if( !::DeleteService( service ) )
	{
		DisplayError( "DeleteService failed" );
	}
	else
	{
		cout << "Service uninstalled!" << endl;
	}

	::CloseServiceHandle( service );
	::CloseServiceHandle( handle );
}

//-----------------------------------------------------------------------------
void CService::Start()
{
	// open a handle to the SCM
	SC_HANDLE handle = ::OpenSCManager( NULL, NULL, SC_MANAGER_ALL_ACCESS );
	if( handle == NULL )
	{
		DisplayError( "Could not connect to SCM dataase" );
		return;
	}

	// open a handle to the service
	SC_HANDLE service = ::OpenService( handle, m_name.c_str(), GENERIC_EXECUTE );
	if( service == NULL )
	{
		::CloseServiceHandle( handle );
		DisplayError( "Could not get handle to service" );
		return;
	}

	// and start the service!
	if( !::StartService( service, 0, NULL ) )
	{
		DisplayError( "StartService failed" );
	}
	else
	{
		cout << "Service started!" << endl;
	}

	::CloseServiceHandle( service );
	::CloseServiceHandle( handle );
}
//-----------------------------------------------------------------------------
void CService::Stop()
{
	// open a handle to the SCM
	SC_HANDLE handle = ::OpenSCManager( NULL, NULL, SC_MANAGER_ALL_ACCESS );
	if( handle == NULL )
	{
		DisplayError( "Could not connect to SCM dataase" );
		return;
	}

	// open a handle to the service
	SC_HANDLE service = ::OpenService( handle, m_name.c_str(), GENERIC_EXECUTE );
	if( service == NULL )
	{
		::CloseServiceHandle( handle );
		return;
	}

	// send the STOP control request to the service
	SERVICE_STATUS status;
	::ControlService( service, SERVICE_CONTROL_STOP, &status );
	::CloseServiceHandle( service );
	::CloseServiceHandle( handle );

	if( status.dwCurrentState != SERVICE_STOPPED )
	{
		DisplayError( "ControlService failed" );
	}
	else
	{
		cout << "Service stopped!" << endl;
	}
}

//-----------------------------------------------------------------------------
void CService::Console( int numArgs, std::string *args )
{
	// we have to make an event object for quitting...
	m_quitEvent = CreateEvent( NULL, FALSE, FALSE, NULL );

	if( !Run( numArgs, args ) )
	{
		Shutdown();
		return;
	}

	HANDLE handles[] = { m_quitEvent, GetStdHandle( STD_INPUT_HANDLE ) };

	// either wait for input on the console, or the quit event...
	while( true )
	{
		if( WaitForMultipleObjects( 2, handles, FALSE, INFINITE ) == (WAIT_OBJECT_0 + 1) )
		{
			INPUT_RECORD record;
			DWORD numEvents;
			memset( &record, 0, sizeof(INPUT_RECORD) );

			// only quit if this is a key event.  Not a mouse event, etc.
			ReadConsoleInput( handles[1], &record, 1, &numEvents );
			if( record.EventType == KEY_EVENT &&
				record.Event.KeyEvent.bKeyDown &&
				record.Event.KeyEvent.wVirtualKeyCode == 27 )	// 27 == escape key
			{
				break;
			}
		}
		else
		{
			break;
		}
	}

	CloseHandle( m_quitEvent );
}

//-----------------------------------------------------------------------------
void CService::ExitService()
{
	if( m_quitEvent )
	{
		SetEvent( m_quitEvent );
	}
	else
	{
		Stop();
	}
}

//-----------------------------------------------------------------------------
void CService::ShowUsage()
{
	cout << "usage: Server.exe action name [options]" << endl << endl;
	cout << "Where: action is one of: -install, -uninstall, -start, -stop, -console" << endl;
	cout << "       name is the name of the service to control." << endl;
	cout << "       options depends on the action given:" << endl;
	cout << "         -install you can give the username and password the service runs under:" << endl;
	cout << "           MyService.exe -install username password" << endl;
	cout << "         -console the arguments you give are passed to the service." << endl;
}

//-----------------------------------------------------------------------------
void CService::DisplayError( std::string const &message )
{
	cout << "error: " << endl;
	cout << message << endl;
}